#!/nix/store/h723hb9m43lybmvfxkk6n7j4v664qy7b-python3-3.11.9/bin/python3

from liquidctl import find_liquidctl_devices
from time import sleep
from time import monotonic

# modules: k10temp nct6775

# /sys/class/hwmon/hwmon1/name is k10temp
# /sys/class/hwmon/hwmon1/temp1_label" is Tctl
CPU_TEMP_FILE = "/sys/class/hwmon/hwmon1/temp1_input"

# /sys/class/hwmon/hwmon2/name is asus
# /sys/class/hwmon/hwmon2/temp1_label" is asusec
CASE_TEMP_FILE = "/sys/class/hwmon/hwmon2/temp3_input"

# /sys/class/hwmon/hwmon5/name is nct6775
RAD_FAN_INPUT_FILE = "/sys/class/hwmon/hwmon5/fan2_input"
RAD_PWM_FILE = "/sys/class/hwmon/hwmon5/pwm2"
CASE_FAN_INPUT_FILE = "/sys/class/hwmon/hwmon5/fan1_input"
CASE_PWM_FILE = "/sys/class/hwmon/hwmon5/pwm1"

RAD_PWM_ENABLE_FILE = RAD_PWM_FILE + "_enable"
CASE_PWM_ENABLE_FILE = CASE_PWM_FILE + "_enable"

# https://github.com/liquidctl/liquidctl/blob/main/docs/kraken-x2-m2-guide.md
# 20 is the actual limit
PUMP_MIN_DUTY = 20

# 60c is absolute max, but CPU may throttle above 40c
LIQUID_BASE_TEMP = 30
LIQUID_MAX_TEMP = 42

CPU_MAX_TEMP = 90
CPU_BASE_TEMP = 48

CASE_MAX_TEMP = 48
CASE_BASE_TEMP = 40

# TODO stall recovery
# TODO zero RPM mode if conditions are met (to avoid hunting)

MANUAL_MODE = 1
SMART_FAN_MODE_4 = 5

UPDATE_INTERVAL_MS = 500


def get_file_int(filepath: str) -> int:
    with open(filepath) as f:
        return int(f.read())


def set_file_int(filepath: str, contents: int) -> None:
    with open(filepath, "w") as f:
        f.write(str(contents))


def min_speed(pwm_file: str, fan_input_file) -> tuple[int, int]:
    # Turn off fans, wait to stop
    set_file_int(pwm_file, 0)

    for x in range(10):
        sleep(0.5)
        if get_file_int(fan_input_file) == 0:
            break
    else:
        raise RuntimeError("Could not stop fan for calibration")

    # ramp up fan, until it starts moving
    for duty in range(100):
        set_file_int(pwm_file, duty)
        sleep(0.5)

        if get_file_int(fan_input_file) > 0:
            break
    else:
        raise RuntimeError("Could not start fan for calibration")

    return duty, get_file_int(fan_input_file)


def max_speed(pwm_file: str, fan_input_file: str) -> int:
    set_file_int(pwm_file, 255)
    sleep(4)
    return get_file_int(fan_input_file)


def set_fan_speed(pwm_file: str, start_duty: int, factor: float):
    """Set fan speed percent 0% == off, 1% -> 100%"""
    factor = max(factor, 0)
    factor = min(factor, 1)
    duty_range = 255 - start_duty
    duty = int(start_duty + factor * duty_range)
    set_file_int(pwm_file, duty)


def stop_fan(pwm_file: str):
    set_file_int(pwm_file, 0)


class Controller:
    def __init__(self):
        for kraken in find_liquidctl_devices():
            if "NZXT Kraken X" in kraken.description:
                self.kraken = kraken
                break
        else:
            raise RuntimeError("Could not find Kraken")

        # from datasheet. < 20 undefined. Calibrate to get actual value
        self.rad_fan_start_duty = int(0.2 * 255)
        self.case_fan_start_duty = int(0.2 * 255)

    def __enter__(self):
        self.kraken.connect()
        self.kraken.initialize()
        set_file_int(RAD_PWM_ENABLE_FILE, MANUAL_MODE)
        set_file_int(CASE_PWM_ENABLE_FILE, MANUAL_MODE)
        return self

    def __exit__(self, type, value, traceback) -> bool:
        # back to "BIOS" control
        set_file_int(RAD_PWM_ENABLE_FILE, SMART_FAN_MODE_4)
        set_file_int(CASE_PWM_ENABLE_FILE, SMART_FAN_MODE_4)
        self.set_pump_speed(0.75)
        self.kraken.disconnect()  # last, might fail

    def calibrate(self):
        # could parallelise, but not worth it with 2 fans
        print("Calibrating start duty...")
        self.rad_fan_start_duty, rad_fan_speed = min_speed(
            RAD_PWM_FILE, RAD_FAN_INPUT_FILE
        )
        self.case_fan_start_duty, case_fan_speed = min_speed(
            CASE_PWM_FILE, CASE_FAN_INPUT_FILE
        )

        print("Rad fan start duty: %s %%" % int(self.rad_fan_start_duty / 2.55))
        print("Rad fan min speed: %s RPM" % rad_fan_speed)
        print("Case fan start duty: %s %%" % int(self.case_fan_start_duty / 2.55))
        print("Case fan min speed: %s RPM" % case_fan_speed)

        # find max RPM for no reason
        print("Calibrating max speed...")
        print("Rad fan max speed: %s RPM" % max_speed(RAD_PWM_FILE, RAD_FAN_INPUT_FILE))
        print(
            "Case fan max speed: %s RPM" % max_speed(CASE_PWM_FILE, CASE_FAN_INPUT_FILE)
        )

    def watch(self):
        last_ts = 0

        while True:
            kraken_status = {s[0]: s[1] for s in self.kraken.get_status()}
            yield {
                "liquid_temp": kraken_status["Liquid temperature"],
                "cpu_temp": get_file_int(CPU_TEMP_FILE) / 1000,
                "case_temp": get_file_int(CASE_TEMP_FILE) / 1000,
                "rad_fan_speed": get_file_int(RAD_FAN_INPUT_FILE),
                "case_fan_speed": get_file_int(CASE_FAN_INPUT_FILE),
                "pump_speed": kraken_status["Pump speed"],
                "pump_duty_percent": kraken_status["Pump duty"],
                "rad_fan_duty_percent": int(get_file_int(RAD_PWM_FILE) / 2.55),
                "case_fan_duty_percent": int(get_file_int(CASE_PWM_FILE) / 2.55),
            }

            # don't drift!
            delta = monotonic() - last_ts
            time_deficit = max(UPDATE_INTERVAL_MS / 1000 - delta, 0)
            sleep(time_deficit)
            last_ts = monotonic()

    def set_pump_speed(self, factor: float):
        """Set pump speed percent, 0 = min speed"""
        factor = max(factor, 0)
        factor = min(factor, 1)
        duty_range = 100 - PUMP_MIN_DUTY
        duty = int(PUMP_MIN_DUTY + factor * duty_range)
        self.kraken.set_fixed_speed("pump", duty)

    def loop(self):
        print("Entering main loop...")
        for status in controller.watch():
            liquid_factor = (status["liquid_temp"] - LIQUID_BASE_TEMP) / (
                LIQUID_MAX_TEMP - LIQUID_BASE_TEMP
            )
            cpu_factor = (status["cpu_temp"] - CPU_BASE_TEMP) / (
                CPU_MAX_TEMP - CPU_BASE_TEMP
            )

            case_factor = (status["case_temp"] - CASE_BASE_TEMP) / (
                CASE_MAX_TEMP - CASE_BASE_TEMP
            )

            controller.set_pump_speed(cpu_factor)
            set_fan_speed(RAD_PWM_FILE, controller.rad_fan_start_duty, liquid_factor)
            set_fan_speed(CASE_PWM_FILE, controller.case_fan_start_duty, case_factor)

            #print(status)


# TODO 3 separate control threads for pump, rad fan and cpu fan
# or, 3 controller class threads with parameters and hwmononly (no liquidctl) -- one day https://github.com/liquidctl/liquidtux
# TODO Experiment with Zero RPM (Hybrid) mode -- if dwell below a threshold for a time, turn fans off until a higher threshold
# TODO avoid beat frequencies between case fan/cpu
# TODO factor into: Subject and Cooler

with Controller() as controller:
    controller.calibrate()
    controller.loop()
